@@ -4,7 +4,7 @@
 #' \code{feglm.nb} can be used to fit negative binomial generalized linear models with many 
 #' high-dimensional fixed effects (see \code{\link{feglm}}).
 #' @param
-#' formula,data,beta.start,eta.start,control see \code{\link{feglm}}.
+#' formula,data,weights,beta.start,eta.start,control see \code{\link{feglm}}.
 #' @param
 #' init.theta an optional initial value for the theta parameter (see \code{\link[MASS]{glm.nb}}).
 #' @param
@@ -19,6 +19,9 @@
 #' Gaure, S. (2013). "OLS with Multiple High Dimensional Category Variables". Computational
 #' Statistics and Data Analysis. 66.
 #' @references 
+#' Marschner, I. (2011). "glm2: Fitting generalized linear models with convergence problems". 
+#' The R Journal, 3(2).
+#' @references 
 #' Stammann, A., F. Heiss, and D. McFadden (2016). "Estimating Fixed Effects Logit Models with Large
 #' Panel Data". Working paper.
 #' @references 
@@ -27,35 +30,38 @@
 #' @seealso
 #' \code{\link[MASS]{glm.nb}}, \code{\link{feglm}} 
 #' @export
-feglm.nb <- function(formula       = NULL,
-                     data          = NULL,
-                     beta.start    = NULL,
-                     eta.start     = NULL,
-                     init.theta    = NULL,
-                     link          = c("log", "identity", "sqrt"),
-                     control       = NULL) {
-  # Validity of input argument (formula)
+feglm.nb <- function(
+  formula       = NULL,
+  data          = NULL,
+  weights       = NULL,
+  beta.start    = NULL,
+  eta.start     = NULL,
+  init.theta    = NULL,
+  link          = c("log", "identity", "sqrt"),
+  control       = NULL
+  ) {
+  # Check validity of formula
   if (is.null(formula)) {
     stop("'formula' has to be specified.", call. = FALSE)
   } else if (!inherits(formula, "formula")) {
     stop("'formula' has to be of class formula.", call. = FALSE)
   }
   
-  # Validity of input argument (data)
+  # Check validity of data
   if (is.null(data)) {
     stop("'data' has to be specified.", call. = FALSE)
   } else if (!inherits(data, "data.frame")) {
     stop("'data' has to be of class data.frame.", call. = FALSE)
   }
   
-  # Validity of input argument (link)
+  # Check validity of link
   link <- match.arg(link)
   
-  # Validity of input argument (control)
+  # Check validity of control
   if (is.null(control)) {
     control <- list()
   } else if (!inherits(control, "list")) {
-    stop("'control' has to be of class list.", call. = FALSE)
+    stop("'control' has to be a list.", call. = FALSE)
   }
   
   # Extract control list
@@ -69,7 +75,7 @@ feglm.nb <- function(formula       = NULL,
   
   # Generate model.frame
   setDT(data)
-  data <- data[, all.vars(formula), with = FALSE]
+  data <- data[, c(all.vars(formula), weights), with = FALSE]
   lhs <- names(data)[[1L]]
   nobs.full <- nrow(data)
   data <- na.omit(data)
@@ -91,8 +97,8 @@ feglm.nb <- function(formula       = NULL,
   
   # Drop observations that do not contribute to the loglikelihood
   if (control[["drop.pc"]]) {
-    for (i in k.vars) {
-      data[, (tmp.var) := mean(get(lhs)), by = eval(i)]
+    for (j in k.vars) {
+      data[, (tmp.var) := mean(get(lhs)), by = eval(j)]
       data <- data[get(tmp.var) > 0.0]
       data[, (tmp.var) := NULL]
     }
@@ -106,21 +112,39 @@ feglm.nb <- function(formula       = NULL,
   }
   
   # Determine number of dropped observations
-  nobs <- c(nobs.full = nobs.full,
-            nobs.na   = nobs.na,
-            nobs.pc   = nobs.full - nrow(data),
-            nobs      = nrow(data))
+  nt <- nrow(data)
+  nobs <- c(
+    nobs.full = nobs.full,
+    nobs.na   = nobs.na,
+    nobs.pc   = nobs.full - nt,
+    nobs      = nt
+  )
   
   # Extract model response and regressor matrix
   y <- data[[1L]]
   X <- model.matrix(formula, data, rhs = 1L)[, - 1L, drop = FALSE]
   nms.sp <- attr(X, "dimnames")[[2L]]
   attr(X, "dimnames") <- NULL
+  p <- ncol(X)
   
   # Check for linear dependence in 'X'
-  p <- ncol(X)
   if (qr(X)[["rank"]] < p) {
-    stop("Linear dependent terms detected!", call. = FALSE)
+    stop("Linear dependent terms detected.", call. = FALSE)
+  }
+  
+  # Extract weights if required
+  if (is.null(weights)) {
+    wt <- rep(1.0, nt)
+  } else {
+    wt <- data[[weights]]
+  }
+  
+  # Check validity of weights
+  if (!is.numeric(wt)) {
+    stop("weights must be numeric.", call. = FALSE)
+  }
+  if (any(wt < 0.0)) {
+    stop("negative weights are not allowed.", call. = FALSE)
   }
   
   # Check starting guess of \theta
@@ -135,6 +159,7 @@ feglm.nb <- function(formula       = NULL,
     }
     family <- negative.binomial(init.theta, link)
   }
+  rm(init.theta)
   
   # Compute and check starting guesses
   if (!is.null(beta.start) || !is.null(eta.start)) {
@@ -156,7 +181,7 @@ feglm.nb <- function(formula       = NULL,
       eta <- as.vector(X %*% beta)
     } else {
       # Validity of input argument (eta.start)
-      if (length(eta.start) != nobs[[4L]]) {
+      if (length(eta.start) != nt) {
         stop("Length of 'eta.start' has to be equal to the number of observations.", call. = FALSE)
       }
       
@@ -167,14 +192,9 @@ feglm.nb <- function(formula       = NULL,
   } else {
     # Compute starting guesses if not user specified
     beta <- numeric(p)
-    eta <- numeric(nobs[[4L]])
-    for (i in k.vars) {
-      data[, (tmp.var) := mean(get(lhs)), by = eval(i)]
-      eta <- eta + family[["linkfun"]](data[[tmp.var]] + 0.1) / k
-      data[, (tmp.var) := NULL]
-    }
+    eta <- rep(family[["linkfun"]](sum(wt * (y + 0.1)) / sum(wt)), nt)
   }
-  rm(beta.start, eta.start, init.theta)
+  rm(beta.start, eta.start)
   
   # Get names and number of levels in each fixed effects category
   nms.fe <- lapply(data[, k.vars, with = FALSE], levels)
@@ -190,29 +210,40 @@ feglm.nb <- function(formula       = NULL,
   trace <- control[["trace"]]
   
   # Initial negative binomial fit
-  fit <- feglmFit(beta, eta, y, X, k.list, family, control)
+  fit <- feglmFit(beta, eta, y, X, wt, k.list, family, control)
   beta <- fit[["coefficients"]]
   eta <- fit[["eta"]]
   dev <- fit[["deviance"]]
-  n <- nobs[[4L]]
-  theta <- suppressWarnings(theta.ml(y, family[["linkinv"]](eta), n = n,
-                                     limit = limit, trace = trace))
+  theta <- suppressWarnings(
+    theta.ml(
+      y     = y,
+      mu    = family[["linkinv"]](eta),
+      n     = nt,
+      limit = limit,
+      trace = trace
+    )
+  )
   
-  
-  # Alternate between fitting a glm and \theta
+  # Alternate between fitting glm and \theta
   conv <- FALSE
   for (iter in seq.int(iter.max)) {
     # Fit negative binomial model
     dev.old <- dev
+    theta.old <- theta
     family <- negative.binomial(theta, link)
-    fit <- feglmFit(beta, eta, y, X, k.list, family, control)
+    fit <- feglmFit(beta, eta, y, X, wt, k.list, family, control)
     beta <- fit[["coefficients"]]
     eta <- fit[["eta"]]
     dev <- fit[["deviance"]]
-    theta0 <- theta
-    theta <- suppressWarnings(theta.ml(y, family[["linkinv"]](eta), n = n,
-                                       limit = limit, trace = trace))
-    
+    theta <- suppressWarnings(
+      theta.ml(
+        y     = y,
+        mu    = family[["linkinv"]](eta),
+        n     = nt,
+        limit = limit,
+        trace = trace
+        )
+      )
     
     # Progress information
     if (trace) {
@@ -223,9 +254,9 @@ feglm.nb <- function(formula       = NULL,
     }
     
     # Check termination condition
-    con1 <- abs(dev - dev.old) / (0.1 + abs(dev))
-    con2 <- abs(theta - theta0)
-    if (con1 + con2 <= tol) {
+    dev.crit <- abs(dev - dev.old) / (0.1 + abs(dev))
+    theta.crit <- abs(theta - theta.old) / (0.1 + abs(theta.old))
+    if (dev.crit <= tol && theta.crit <= tol) {
       if (trace) {
         cat("Convergence\n")
       }
@@ -235,21 +266,27 @@ feglm.nb <- function(formula       = NULL,
   }
   rm(y, X, eta)
   
-  # Add names to \beta, Scores, and Hessian
+  # Information if convergence failed
+  if (!conv && trace) cat("Algorithm did not converge.\n") 
+  
+  # Add names to \beta, Hessian, and MX (if provided)
   names(fit[["coefficients"]]) <- nms.sp
-  colnames(fit[["Score"]]) <- nms.sp
+  if (control[["keep.mx"]]) {
+    colnames(fit[["MX"]]) <- nms.sp
+  }
   dimnames(fit[["Hessian"]]) <- list(nms.sp, nms.sp)
   
   # Return result list
-  structure(c(fit, list(theta      = theta,
-                        iter.outer = iter,
-                        conv.outer = conv,
-                        nobs       = nobs,
-                        lvls.k     = lvls.k,
-                        nms.fe     = nms.fe,
-                        formula    = formula,
-                        data       = data,
-                        family     = family,
-                        control    = control)),
-            class = c("feglm", "feglm.nb"))
+  structure(c(fit, list(
+    theta      = theta,
+    iter.outer = iter,
+    conv.outer = conv,
+    nobs       = nobs,
+    lvls.k     = lvls.k,
+    nms.fe     = nms.fe,
+    formula    = formula,
+    data       = data,
+    family     = family,
+    control    = control)
+    ), class = c("feglm", "feglm.nb"))
 }
\ No newline at end of file
